#################################################################################
# Created       : Jeremy D. Zawodny
# Modified      : 08/11/2001
#################################################################################
#
# Weather - Module part of Webchat designed to handle the weather.
# Copyright (C) 1997-1998  Jeremy D. Zawodny
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : jzawodn@pobox.com
#
#################################################################################

package Ethereal::Weather;					# Pacakace name

use strict;							# Strict variable enforcement
use LWP::UserAgent;						# WWW access in Perl

#################################################################################
# Sub-Routines
#################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	my $class = shift;
	my $self = {};
	bless $self, $class;					# Bless $self into designated class
	return $self;						
}


#####################
# Parse Weather
#
# This is a long and involved subroutine. It basically copies the @tokens
# array and treats it as a stack, popping off items, examining them, and
# see what they look like.  Based on their 'apppearance' it takes care
# populating the proper fields internally.

sub Parse 
{

	#####################
	# Data Members

	my $self   = shift;					# Self reference
	my $tok    = shift;					# List of tokens in reading
	my $fields = shift;					# Reference to a hash

	my $cond;						# Individual condition

	my @toks;						# Series of tokens
	my %cond;						# Weather conditions


	#####################
	# Program Area

	# Initial initialization
	# List values in fields
	%{$fields} = (
			WEATHER	=> [ ],				# Weather for export
			HUMAN	=> [ ],				# Human readable
			SKY	=> [ ],				# Sky conditions for export
			weather => [ ],				# Internal weather
			sky	=> [ ],				# Internal sky
			remarks => [ ]				# Internal remarks
		);	

	# Weather conditions
	%cond = (
			VC	=> 'In vicinity of ',
			MI	=> 'Shallow ',
			BC	=> 'Patches ',
			PR	=> 'Partial ',
			TS	=> 'Thunderstorm ',
			BL	=> 'Blowing ',
			SH	=> 'Showers ',
			DR	=> 'Drizzle ',
			FZ	=> 'Freezing ',
			DZ	=> 'Drizzle',
			RA	=> 'Rain',
			SN	=> 'Snow',
			SG	=> 'Snow grains',
			IC	=> 'Ice crystals',
			PL	=> 'Ice pellets',
			GR	=> 'Hail',
			GS	=> 'Small hail/pellets',
			BR	=> 'Mist',
			FG	=> 'Fog',
			FU	=> 'Smoke',
			VA	=> 'Volcanic Ash',
			SA	=> 'Sand',
			HZ	=> 'Haze',
			PY	=> 'Spray',
			DU	=> 'Widespead dust',
			SQ	=> 'Squall',
			SS	=> 'Sandstorm',
			DS	=> 'Duststorm',
			PO	=> 'Well developed dust/sand whirls',
			FC	=> 'Funnel cloud',
			+FC	=> 'Tornado/waterspout'
		);

	# Retreive tokens
	@toks = split(/\s+/, uc($tok));


	# as long as there are tokens
	while($tok = shift(@toks)) 
	{
		# Report type
    		if (($tok =~ /METAR/i) or ($tok =~ /SPECI/i)) 
		{
      			$fields->{type} = $tok;
      			next;
		}

      		# Site ID
		elsif ($tok =~ /K[A-Z]{3,3}/) 
		{	
			$fields->{site} = $tok;
			next;
		}

		# Date/time
		elsif($tok =~ /\d{6,6}Z/i) 
		{
			$fields->{date_time} = $tok;
			next;
		}

		# Report modifier
		elsif(($tok =~ /AUTO/i) or ($tok =~ /COR/i)) 
		{
			$fields->{modifier} = $tok;
			next;
		}

		# Wind information
		elsif($tok =~ /.*?KT$/i) 
		{
			$fields->{wind} = $tok;
			next;
		}

		# Visibility information
		elsif($tok =~ /.*?SM$/i) 
		{
			$fields->{visibility} = $tok;
			next;
		}

		# Visibility information with a leading digit
      		elsif($tok =~ /^\d$/) 
		{

			$tok .= " " . shift(@toks);
			$fields->{visibility} = $tok;
			next;
		}

		# Runway visibility info
		elsif($tok =~ /R.*?FT$/i)
		{
			$fields->{runway} = $tok;
			next;
		}

		# Current weather info
      		elsif($tok =~ /^(-|\+|VC)?(TS|SH|FZ|BL|DR|MI|BC|PR|RA|DZ|SN|SG|GR|GS|PE|IC|UP|BR|FG|FU|VA|DU|SA|HZ|PY|PO|SQ|FC|SS|DS)+$/)
		{
			push(@{$fields->{weather}},$tok);
			next;
		}

		# Sky conditions (clouds)
		elsif(($tok =~ /SKC|CLR/i) or ($tok =~ /(FEW|SCT|BKN|OVC)(\d\d\d)(CB|TCU)?$/i))
		{
			push(@{$fields->{sky}},$tok);
			next;
		}

		# Temperature and dew point info
      		elsif($tok =~ /(M?\d\d)\/(M?\d\d)/i)
		{
			$fields->{temp_dew} = $tok;
			next;
		}

		# Altimeter setting?
       		elsif($tok =~ /A\d{4,4}$/i) 
		{
			$fields->{alt} = $tok;
			next;
		}

		# Assume remarks
		else 
		{
			push(@{$fields->{remarks}},$tok);
			next;
      		}

  	} 

	# Now that the internal stuff is set, let's do the external stuff.
	#

	# Type
	if ($fields->{type} eq "METAR") 
	{
    		$fields->{TYPE} = "Routine Weather Report";
  	}

	# Misc
	$fields->{SITE} = $fields->{site};
	$fields->{DATE} = substr($fields->{date_time},0,2);
	$fields->{TIME} = substr($fields->{date_time},2,4) . " UTC";
	$fields->{TIME} =~ s/(\d\d)(\d\d)/$1:$2/o;
	$fields->{MOD}  = $fields->{modifier};


	# Okay, wind finally gets interesting.
	{
    		my $wind = $fields->{wind};
    		my $dir_deg  = substr($wind,0,3);
    		my $dir_eng = "";
	
    		# Check for wind direction
		# Variable
		if ($dir_deg =~ /VRB/i)
		{
			$dir_deg = "Variable";
    		} 

		# Directional
		else 
		{
      			if ($dir_deg < 15)     { $dir_eng = "North"; } 
			elsif ($dir_deg < 30)  { $dir_eng = "North/Northeast"; } 
			elsif ($dir_deg < 60)  { $dir_eng = "Northeast"; } 
			elsif ($dir_deg < 75)  { $dir_eng = "East/Northeast"; } 
			elsif ($dir_deg < 105) { $dir_eng = "East"; } 
			elsif ($dir_deg < 120) { $dir_eng = "East/Southeast"; } 
			elsif ($dir_deg < 150) { $dir_eng = "Southeast"; } 
			elsif ($dir_deg < 165) { $dir_eng = "South/Southeast"; } 
			elsif ($dir_deg < 195) { $dir_eng = "South"; } 
			elsif ($dir_deg < 210) { $dir_eng = "South/Southeast"; } 
			elsif ($dir_deg < 240) { $dir_eng = "Southwest"; } 
			elsif ($dir_deg < 265) { $dir_eng = "South/Southwest"; } 
			elsif ($dir_deg < 285) { $dir_eng = "West"; } 
			elsif ($dir_deg < 300) { $dir_eng = "West/Northwest"; } 
			elsif ($dir_deg < 330) { $dir_eng = "Northwest"; } 
			elsif ($dir_deg < 345) { $dir_eng = "North/Northwest"; } 
			else { $dir_eng = "North"; }
		}
	
    		$wind =~ /...(\d\d\d?)/o;
    		my $kts_speed = $1;
    		my $mph_speed = $kts_speed * 1.1508;
		my $kph_speed = $mph_speed * 1.61;
    		my $kts_gust = 0;
    		my $mph_gust = 0;
		my $kph_gust = 0;
	
    		if ($wind =~ /.{5,6}G(\d\d\d?)/o)
		{
			$kts_gust = $1;
      			$mph_gust = $kts_gust * 1.1508;
			$kph_gust = $mph_speed * 1.61;   
 		}

		$fields->{WIND_KTS} = $kts_speed;
		$fields->{WIND_MPH} = sprintf("%.1f", $mph_speed);
		$fields->{WIND_KPH} = sprintf("%.1f", $kph_speed);
	
		$fields->{WIND_KTS_GUST} = $kts_gust;
		$fields->{WIND_MPH_GUST} = sprintf("%.1f", $mph_gust);
		$fields->{WIND_KPH_GUST} = sprintf("%.1f", $kph_gust);

		$fields->{WIND_DIR_DEG} = $dir_deg;
		$fields->{WIND_DIR_ENG} = $dir_eng;

	} # End wind block


	# Visibility, now.
	{
		my $vis = $fields->{visibility};
    		$vis =~ s/SM$//oi;

		if ($vis =~ /M(\d\/\d)/o) { $fields->{VISIBILITY} = "Less than $1 statute miles"; } 
		else { $fields->{VISIBILITY} = $vis . " Statute Miles"; }

		# Alternates
		$fields->{VIS_MILE} = $vis;
		$fields->{VIS_KILO} = $vis * 1.61;

	} # End visibility block


	# And F/C temperatures.
	{
    		my ($tmp,$dew) = split(/\//, $fields->{temp_dew});

		# check for negative values
		$tmp =~ s/^M/-/o;
		$dew =~ s/^M/-/o;

		# convert celcius to farenheit
		$fields->{C_TEMP} = $tmp;
		$fields->{F_TEMP} = (($tmp * (9/5)) + 32);
		$fields->{C_DEW} = $dew;
		$fields->{F_DEW} = (($dew * (9/5)) + 32);
  	}


	# Barometric pressure
	{
		# Basic findings
		my $bar = substr($fields->{alt},1);
		$bar = $bar / 100;

		# Export
		# Inches of mercury
		$fields->{ALTI} = $bar;

		# Kilopascals
		$fields->{ALTK} = sprintf("%.2f", ($bar * 3.3864));
	}


	# Cloud Cover
	{
		my $alt;

		# Cycle through fields
		while ($tok = shift(@{$fields->{sky}}))
		{
			# Determine and work with
			if ($tok =~ /CLR/i) { $tok = 'Clear'; }
			elsif ($tok =~ /FEW(\d\d\d)(CB|TCU)?$/i) { $tok = 'Few clouds'; } 
			elsif ($tok =~ /SCT(\d\d\d)(CB|TCU)?$/i) { $tok = 'Scattered clouds'; } 
			elsif ($tok =~ /BKN(\d\d\d)(CB|TCU)?$/i) { $tok = 'Broken clouds'; } 
			elsif ($tok =~ /OVC(\d\d\d)(CB|TCU)?$/i) { $tok = 'Overcast clouds'; } 

			# Calculate altitude
			$alt = $1 * 100;			

			# Export value
			if ($2 =~ /CB/) { push(@{$fields->{SKY}}, "$tok (Cumulonimbus) at $alt feet"); } 
			elsif ($2 =~ /TCU/) { push(@{$fields->{SKY}}, "$tok (Towering cumulus) at $alt feet"); }
			else { push(@{$fields->{SKY}}, "$tok at $alt feet"); }
		}
	}


	# Weather Patterns
	{
		# Cycle through weather
		while ($tok = shift(@{$fields->{weather}}))
		{
			# Send normal code
			push(@{$fields->{WEATHER}}, $tok);


			# Cycle through keys
			foreach $cond (keys(%cond))
			{
				# Replace
				$tok =~ s/$cond/$cond{$cond}/g;
			}

			# Search for + or -
			$tok =~ s/\-/Light /g;
			$tok =~ s/\+/Heavy /g;

			
			# Add to human readable
			push(@{$fields->{HUMAN}}, $tok);
		}
	}
}


#####################
# Weather Get
#
# Will make use of LWP::UserAgent in order to retrieve the information
# from http://weather.noaa.gov/cgi-bin/mgetmetar.pl and get the needed
# METAR Observation(s) from that returned document.

sub Get
{

	#####################
	# Data members

	my $self   = shift;					# Self reference

	my $cnt;						# HTTP Contents	
	my $req;						# HTTP Request
	my $res;						# HTTP Response
	my $ua;							# User agent

	my $par;						# Parameters
	my $site = 'http://weather.noaa.gov/cgi-bin/mgetmetar.pl';

	my @locale = @_;					# List of locales
	my %locale;						# List of locales and readings

	#####################
	# Program Area

	# Generate parameter list
	$par = join("%20", @locale);

	# Retreival
	# Create user agent
	$ua = new LWP::UserAgent;
	$ua->agent("AgentName/0.1 " . $ua->agent);

	# Create a request
	$req = new HTTP::Request POST => $site;
	$req->content_type('application/x-www-form-urlencoded');
	$req->content("cccc=$par");		 	

	# Pass over to user agent
	$res = $ua->request($req);

	# Check the outcome of the response	
	if ($res->is_success)
	{
		# Addtional variables
		my $loc;					# Locale
		my $str;					# Temporary string


		# Capture contents
		$cnt = $res->content;

		# Find strings
		foreach $loc (@locale)	
		{
			# Find string
			if ($cnt =~ /($loc \d{6})(.*?)<\/FONT>/is)
			{
				# Assign and cleanup
				$str = "$1$2";
				$str =~ s/\s+/ /gi;

				# Initialize hash value
				$locale{$loc} = $str;	
			}	
		}
	}

	# Return hash
	return %locale;
}
1;								# Return True
